#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>


int ***alloc3d_int(int n, int m, int k) {
    int *data=(int *)malloc(n*m*k*sizeof(int));
    int ***array = (int ***)malloc(n*sizeof(int **));

    for (int i=0; i<n; i++) {
        array[i] = (int **)malloc(m*sizeof(int *));
        for (int j=0; j<m; j++) {
            array[i][j] = &(data[(i*m+j)*k]);
        }
    } 
    return array; 
}

void free3d_int(int ***array, int n, int m, int k) {
    free(array[0][0]);
    for (int i=0; i<n; i++)
        free(array[i]);
    free(array);
}


int **alloc2d_int(int n, int m) {
    int *data=(int *)malloc(n*m*sizeof(int));
    int **array = (int **)malloc(n*sizeof(int *));

    for (int i=0; i<n; i++) {
        array[i] = &(data[i*m]);
    } 
    return array; 
}

void free2d_int(int **array, int n, int m) {
    free(array[0]);
    free(array);
}

int main(int argc, char **argv) {
    int ierr, rank, size;
    int n=1, m=1;
    int ***randdata=NULL;
    int **results=NULL;

    ierr = MPI_Init(&argc, &argv);
    ierr|= MPI_Comm_size(MPI_COMM_WORLD,&size);
    ierr|= MPI_Comm_rank(MPI_COMM_WORLD,&rank);

    if (argc != 3) {
        fprintf(stderr,"Usage: %s N M, where N,M  are sizes of arrays.\n",
                argv[0]);
        exit(-1);
    }

    n = atoi(argv[1]);
    m = atoi(argv[2]);
    srand(rank);
    randdata = alloc3d_int(n,m,100);
    printf("Rank %d: starting to generate data [%dx%d].\n",rank,n,m);
    for (int i=0; i<n; i++) {
        for (int j=0; j<m; j++) {
            for (int k=0; k<100; k++) {
                randdata[i][j][k] = rand();
            }
        }
    } 


    /* rank 0 gets all the maxima */
    if (rank == 0) {
        results = alloc2d_int(n,m);
    }
    printf("Rank %d done; now finding local max of each row,col.\n",rank); 
    for (int i=0; i<n; i++) {
        for (int j=0; j<m; j++) {
            int locmax = -9999999;
            for (int k=0; k<100; k++) 
                if (randdata[i][j][k] > locmax) locmax = randdata[i][j][k];

            MPI_Reduce(&locmax, &(results[i][j]), 1, MPI_INT, MPI_MAX, 0, MPI_COMM_WORLD);

        }
    } 
    if (rank == 0) {
        for (int i=0; i<n; i++) {
            for (int j=0; j<m; i++) {
                printf("Maximum in (%d,%d) - %d\n", i, j, results[i][j]);
            } 
        } 
    }
    

    if (rank == 0) 
        free2d_int(results,n,m);

    free3d_int(randdata, n, m, 100);
    MPI_Finalize();
    return 0;
}
