#include <stdio.h>
#include <omp.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "../util/pca_utils.h"


int main(int argc, char *argv[])
{
  assert(argc>1);
  int n=atoi(argv[1]);
  double **mat=matrix(n,n);
  memset(mat[0],0,sizeof(double)*n*n);

  omp_lock_t *locks=(omp_lock_t *)malloc(sizeof(omp_lock_t)*n);
  for (int i=0;i<n;i++)
    omp_init_lock(&locks[i]);
  
  pca_time tt;

  tick(&tt);
#pragma omp parallel shared(mat,n,locks) default(none)
  {
    double **mymat=matrix(n,n);
    int myid=omp_get_thread_num();
    for (int i=0;i<n;i++)
      for (int j=0;j<n;j++)
	mymat[i][j]=myid;    
#pragma omp critical
    for (int i=0;i<n;i++)
      for (int j=0;j<n;j++)
	mat[i][j]+=mymat[i][j];
    free(mymat[0]);
    free(mymat);
  }
  printf("took %12.4g seconds to fill and reduce critical, first element is %8.3f\n",tocksilent(&tt),mat[0][0]);
  memset(mat[0],0,sizeof(double)*n*n);


  tick(&tt);
#pragma omp parallel shared(mat,n,locks) default(none)
  {
    double **mymat=matrix(n,n);
    int myid=omp_get_thread_num();
    for (int i=0;i<n;i++)
      for (int j=0;j<n;j++)
	mymat[i][j]=myid;    

    for (int i=0;i<n;i++) {
      omp_set_lock(&locks[i]);
      for (int j=0;j<n;j++)
	mat[i][j]+=mymat[i][j];
      omp_unset_lock(&locks[i]);
    }
    free(mymat[0]);
    free(mymat);
  }
  printf("took %12.4g seconds to fill and reduce locks, first element is %8.3f\n",tocksilent(&tt),mat[0][0]);
  
  

}
