#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#ifdef PGPLOT
#include <cpgplot.h>
#endif
#include <omp.h>
#include <complex.h>
#include "../util/pca_utils.h"

#include "mandel.h"

#define AUTOSCALE -123456

/*--------------------------------------------------------------------------------*/

char display_map_new_limits(float **map, int npix, int fmin, int fmax, double ind, double *xmin, double *xmax, double *ymin, double *ymax)
{
#ifndef PGPLOT
  exit(EXIT_SUCCESS);
#else
  //float tr[6] = {0.0, 1.0, 0.0, 0.0, 0.0, 1.0};
  static int firsttime=1;
  if (firsttime) {
    cpgopen("?");
    cpgenv(1,npix,1,npix,1,-2);  
    my_pgctab(0,1,0,1,1,0,256);
    firsttime=0;
  }

  float tr[6] = {0.0, 1.0, 0.0, 0.0, 0.0, 1.0};
  if (fmin==AUTOSCALE){
    fmin=map[0][0];
    for (int i=0;i<npix;i++)
      for (int j=0;j<npix;j++)
	if (map[i][j]<fmin)
	  fmin=map[i][j];
  }
  
  
  if (fmax==AUTOSCALE){
    fmax=map[0][0];
    for (int i=0;i<npix;i++)
      for (int j=0;j<npix;j++)
	if (map[i][j]>fmax)
	  fmax=map[i][j];
  }

  float **map_copy=fmatrix(npix,npix);
  for (int i=0;i<npix;i++)
    for (int j=0;j<npix;j++)
      map_copy[i][j]=pow(map[j][i],ind); //take transpose for plotting purposes
  fmin=pow(fmin,ind);
  fmax=pow(fmax,ind);

  cpgimag(map_copy[0],npix,npix,1,npix,1,npix,fmin,fmax,tr);

  free_fmatrix(map_copy);
  

  float fx1in, fy1in, fx2in, fy2in;
  char press;
  cpgcurs(&fx1in,&fy1in,&press);
  if (press=='X') {
    *xmin=(fx1in/npix)*(*xmax-*xmin)+(*xmin);
    *ymin=(fy1in/npix)*(*ymax-*ymin)+(*ymin);
    return press;
  }
  cpgsci(3);  
  cpgband(2,0,fx1in,fy1in,&fx2in,&fy2in,&press);


  double dfx1in=(fx1in/npix)*(*xmax-*xmin)+(*xmin);
  double dfx2in=(fx2in/npix)*(*xmax-*xmin)+(*xmin);
  double dfy1in=(fy1in/npix)*(*ymax-*ymin)+(*ymin);
  double dfy2in=(fy2in/npix)*(*ymax-*ymin)+(*ymin);

  double dx=fabs(dfx2in-dfx1in);
  double dy=fabs(dfy2in-dfy1in);
  if (dy>dx)
    dx=dy;
  double xcent=0.5*(dfx1in+dfx2in);
  double ycent=0.5*(dfy1in+dfy2in);

  *xmin=xcent-0.5*dx;
  *xmax=*xmin+dx;
  *ymin=ycent-0.5*dx;
  *ymax=*ymin+dx;

  return press;

#endif
}

/*--------------------------------------------------------------------------------*/

int how_many_iter(double complex a, int maxiter)
     /*At complex point a, figure out how many iterations it takes
       to know the point is outside of the Mandlebrot set.  Or, return maxiter
       if (as far as we know) the point is inside the set.*/
{
  int i=0;
  double complex b=a;
  double r;
  while (i<maxiter) {
    r=cimag(b)*cimag(b)+creal(b)*creal(b);
    if (r>4)
      return i;
    b=b*b+a;
    i++;
  }
  return maxiter;
}


/*--------------------------------------------------------------------------------*/

int how_many_julia_iter(double complex a, int maxiter, double complex val)
{
  int i=0;
  double complex b=a;
  double r;
  while (i<maxiter) {
    r=cimag(b)*cimag(b)+creal(b)*creal(b);
    if (r>4)
      return i;
    b=b*b+val;
    i++;
  }
  return maxiter;
}

/*--------------------------------------------------------------------------------*/
void my_pgctab(float r,float R, float g, float G, float b, float B, int nc)
{
#ifndef PGPLOT
  return;
#else
  float *rr=(float *)malloc(sizeof(float)*nc);
  float *gg=(float *)malloc(sizeof(float)*nc);
  float *bb=(float *)malloc(sizeof(float)*nc);
  float *ll=(float *)malloc(sizeof(float)*nc);
  for (int i=0;i<nc;i++) {
    rr[i]=r+(R-r)*(float)i/(float)(nc-1);
    gg[i]=g+(G-g)*(float)i/(float)(nc-1);
    bb[i]=b+(B-b)*(float)i/(float)(nc-1);
    ll[i]=(float)i/(float)(nc-1);
  }

  cpgctab(ll,rr,gg,bb,nc,1.0,0.5);

  free(rr);
  free(gg);
  free(bb);
  free(ll);
#endif    
}
/*--------------------------------------------------------------------------------*/
float **make_mandel_map_static(double xmin, double xmax, double ymin, double ymax, int npix, int maxiter)
{
  float **mymap=fmatrix(npix,npix);
#pragma omp parallel for shared(mymap,npix,xmin,xmax,ymin,ymax,maxiter) default(none)  schedule(static,4)
  for (int i=0;i<npix;i++)
    for (int j=0;j<npix;j++) {
      double x=((double)i)/((double)npix)*(xmax-xmin)+xmin;
      double y=((double)j)/((double)npix)*(ymax-ymin)+ymin;
      double complex a=x+I*y;
      mymap[i][j]=how_many_iter(a,maxiter);
    }
  return mymap; 
}
/*--------------------------------------------------------------------------------*/
float **make_mandel_map_dynamic(double xmin, double xmax, double ymin, double ymax, int npix, int maxiter)
{
  float **mymap=fmatrix(npix,npix);
#pragma omp parallel for shared(mymap,npix,xmin,xmax,ymin,ymax,maxiter) default(none)  schedule(dynamic,4)
  for (int i=0;i<npix;i++)
    for (int j=0;j<npix;j++) {
      double x=((double)i)/((double)npix)*(xmax-xmin)+xmin;
      double y=((double)j)/((double)npix)*(ymax-ymin)+ymin;
      double complex a=x+I*y;
      mymap[i][j]=how_many_iter(a,maxiter);
    }
  return mymap; 
}

/*--------------------------------------------------------------------------------*/
float **make_mandel_map_omp(double xmin, double xmax, double ymin, double ymax, int npix, int maxiter)
{
  float **mymap=fmatrix(npix,npix);
#pragma omp parallel for shared(mymap,npix,xmin,xmax,ymin,ymax,maxiter) default(none)  
  for (int i=0;i<npix;i++)
    for (int j=0;j<npix;j++) {
      double x=((double)i)/((double)npix)*(xmax-xmin)+xmin;
      double y=((double)j)/((double)npix)*(ymax-ymin)+ymin;
      double complex a=x+I*y;
      mymap[i][j]=how_many_iter(a,maxiter);
    }
  return mymap; 
}

/*--------------------------------------------------------------------------------*/
float **make_mandel_map_scalar(double xmin, double xmax, double ymin, double ymax, int npix, int maxiter)
{
  float **mymap=fmatrix(npix,npix);
  for (int i=0;i<npix;i++)
    for (int j=0;j<npix;j++) {
      double x=((double)i)/((double)npix)*(xmax-xmin)+xmin;
      double y=((double)j)/((double)npix)*(ymax-ymin)+ymin;
      double complex a=x+I*y;
      mymap[i][j]=how_many_iter(a,maxiter);
    }
  return mymap; 
}

/*--------------------------------------------------------------------------------*/
float **make_julia_map(double xmin, double xmax, double ymin, double ymax, int npix, int maxiter, double complex val)
{
  float **mymap=fmatrix(npix,npix);
#pragma omp parallel for shared(mymap,npix,xmin,xmax,ymin,ymax,maxiter,val) default(none)  schedule(static,4)
  for (int i=0;i<npix;i++)
    for (int j=0;j<npix;j++) {
      double x=((double)i)/((double)npix)*(xmax-xmin)+xmin;
      double y=((double)j)/((double)npix)*(ymax-ymin)+ymin;
      double complex a=x+I*y;
      mymap[i][j]=how_many_julia_iter(a,maxiter,val);
    }
  return mymap; 
}

/*================================================================================*/


int main(int argc, char *argv[])
{
  pca_time tt;
  int npix=800;
  int maxiter=1000;

  double xmin=-2;
  double xmax=2;
  double ymin=-2;
  double ymax=2;


  char pressed=' ';
  double complex julia_val;
  //display_map(map,npix,AUTOSCALE,AUTOSCALE,0.3);
  while (pressed!='X') {
    tick(&tt);
    float **map=make_mandel_map_dynamic(xmin, xmax,ymin,ymax,npix,maxiter);
    tock(&tt);
    pressed=display_map_new_limits(map,npix,AUTOSCALE,AUTOSCALE,0.3,&xmin,&xmax,&ymin,&ymax);
    julia_val=xmin+I*ymin;
    free_fmatrix(map);
  }
  pressed=' ';
  fprintf(stderr,"Making Julia set with values %14.10f %14.10f\n",creal(julia_val),cimag(julia_val));
  xmin=-2;xmax=2;ymin=-2;ymax=2;
  while (pressed !='X'){
    tick(&tt);
    float **map=make_julia_map(xmin, xmax,ymin,ymax,npix,maxiter,julia_val);
    tock(&tt);
    pressed=display_map_new_limits(map,npix,AUTOSCALE,AUTOSCALE,0.3,&xmin,&xmax,&ymin,&ymax);
    free_fmatrix(map);
    
  }
  
  
}

