#include <stdio.h>
#include <stdlib.h>
#include <omp.h>
#include <assert.h>
#include <math.h>
#include "../util/pca_utils.h"

/*--------------------------------------------------------------------------------*/
void print_mat(double **mat, int n)
{
  for (int i=0;i<n;i++)
    {
      for (int j=0;j<n;j++)
	fprintf(stdout," %14.5e",mat[i][j]);
      fprintf(stdout,"\n");
    }
}

/*--------------------------------------------------------------------------------*/


void fill_random_matrix(double **mat, int n)
{
  for (int i=0;i<n;i++)
    for (int j=0;j<n;j++)
      mat[i][j]=((double)rand())/RAND_MAX;
}
/*--------------------------------------------------------------------------------*/

void matmult_slowest(double **a, double **b, double **c, int n)
{
  for (int i=0;i<n;i++)
    for (int j=0;j<n;j++)
      {
	c[i][j]=0;
	for (int k=0;k<n;k++)
	  c[i][j]+=a[i][k]*b[k][j];
      }
}
/*--------------------------------------------------------------------------------*/
double matrix_sum(double **a, int n)
{
  double sum=0;
  for (int i=0;i<n;i++)
    for (int j=0;j<n;j++)
      sum+=a[i][j];
  return sum;
}

/*================================================================================*/

int main(int argc, char *argv[])
{
  pca_time tt;
  int n=1000;

  double **a=matrix(n,n);
  double **b=matrix(n,n);
  double **c=matrix(n,n);

  fill_random_matrix(a,n);
  fill_random_matrix(b,n);

  tick(&tt);
  matmult_slowest(a,b,c,n);
  printf("Time to multiply %d x %d matrices with slow multiplication is %9.4f\n",n,n,tocksilent(&tt));
  printf("Sum of elements is %14.4e\n",matrix_sum(c,n));

}
