#include <stdio.h>
#include <omp.h>
#include <stdlib.h>
#include <assert.h>
#include "../util/pca_utils.h"

#define PAUSE_LEN 0.1

void dgemm_(char *transa, char *transb, int *m, int *n, int *k, double *alpha, double *a, int *lda, double *b, int *ldb,double *begta, double *c, int *ldc, int *info);


void dgemm(char transa, char transb, int m, int n, int k, double alpha, double *a, int lda, double *b, int ldb,double beta, double *c, int ldc, int *info)
{
  dgemm_(&transa,&transb,&m,&n,&k,&alpha,a,&lda,b,&ldb,&beta,c,&ldc,info);
}

/*--------------------------------------------------------------------------------*/
void dgemm_square(double *a, double *b, double *c, int n)
{
  int info=0;
  dgemm('N','N',n,n,n,1.0,a,n,b,n,0.0,c,n,&info);
}
/*--------------------------------------------------------------------------------*/

void fill_matrix(double **mat, int n, int offset)
{
  for (int i=0;i<n;i++)
    for (int j=0;j<n;j++)
      mat[i][j]=i+j+offset;
  //add a pause in here, simulating a read from disk
  double t=omp_get_wtime();
  while (omp_get_wtime()-t<PAUSE_LEN)
    ;
}

/*--------------------------------------------------------------------------------*/


int main(int argc, char *argv[])
{
  assert(argc>2);
  int n=atoi(argv[1]);
  int niter=atoi(argv[2]);

  double **a[2];
  double **b[2];
  double **c[2];

  a[0]=matrix(n,n);
  a[1]=matrix(n,n);
  b[0]=matrix(n,n);
  b[1]=matrix(n,n);
  c[0]=matrix(n,n);
  c[1]=matrix(n,n);

#pragma omp parallel shared(a,b,c,n,niter) default(none)
#pragma omp single
  {
    int curset=0;
    int offset=0;
    
    fill_matrix(a[0],n,0);
    fill_matrix(a[0],n,1);

    for (int i=0;i<niter;i++) {
#pragma omp task 
      {
	if (i<niter-1) { //we're one step ahead here in the fill, so skip the last loop iteration
	  printf("filling matrices %d.\n",i+1);
	  fill_matrix(a[1-offset],n,i+1);
	  fill_matrix(b[1-offset],n,i+2);
	  printf("filled matrices %d.\n",i+1);
	}
      }
#pragma omp task 
      {
	printf("multiplying matrices %d\n",i);
	dgemm_square(a[offset][0],b[offset][0],c[offset][0],n);
	printf("multiplied matrices %d\n",i);
      }
#pragma omp taskwait
      offset=1-offset;
      printf("\n");
    }
  }
}
