#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include "../util/pca_utils.h"

/*-------------------------------------------------------------------------*/

//void dgemm_(char *transa, char *transb, int *m, int *n, int *k, double  *alpha, double *a, int *lda, double *b, int *ldb, double *beta,double *c, int *ldc, int alen, int blen);

//void gemm(char transa, char transb, int m, int n, int k, double alpha, double *a, int lda, double *b, int ldb, double beta, double *c, int ldc)
//{
//    dgemm_(&transa, &transb, &m, &n, &k, &alpha, a, &lda, b, &ldb, &beta, c, &ldc, 1, 1);
//}

//
//void matmult_blas(double **a, double **b, double **c, int n) 
//{
//  gemm('N','N',n,n,n,1.0,b[0],n,a[0],n,0.0,c[0],n);
//}


/*-------------------------------------------------------------------------*/
void print_mat(double **mat, int n)
{
  for (int i=0;i<n;i++)
    {
      for (int j=0;j<n;j++)
	fprintf(stdout," %14.5e",mat[i][j]);
      fprintf(stdout,"\n");
    }
}

/*-------------------------------------------------------------------------*/
void fill_random_matrix(double **mat, int n)
{
  for (int i=0;i<n;i++)
    for (int j=0;j<n;j++)
      mat[i][j]=((double)rand())/RAND_MAX;
}

/*-------------------------------------------------------------------------*/
void matrix_transpose(double **in, double **out, int n)
{
  for (int i=0;i<n;i++)
    for (int j=0;j<n;j++)
      out[j][i]=in[i][j];
}

/*-------------------------------------------------------------------------*/
double find_matrix_error(double **a, double **b, int n)
{
  double sum=0;
  for (int i=0;i<n;i++)
    for (int j=0;j<n;j++)
      sum += fabs(a[i][j]-b[i][j]);
  return sum;
}

/*-------------------------------------------------------------------------*/
void matmult_slowest(double **a, double **b, double **c, int n)
{
#pragma omp parallel for default(none) shared(a,b,c,n) 
  for (int i=0;i<n;i++)
    for (int j=0;j<n;j++)
      {
         c[i][j]=0;
         for (int k=0;k<n;k++)
            c[i][j]+=a[i][k]*b[k][j];
      }
}

/*-------------------------------------------------------------------------*/

void matmult_faster(double **a, double **b, double **c, int n)
{
  double **bt=matrix(n,n);
  matrix_transpose(b,bt,n);
#pragma omp parallel for shared(a,bt,c,n)
  for (int i=0;i<n;i++)
    for (int j=0;j<n;j++)
      {
         c[i][j]=0;
         for (int k=0;k<n;k++)
            c[i][j]+=a[i][k]*bt[j][k];
      }
  free_matrix(bt);
}

/*-------------------------------------------------------------------------*/
void matmult_block(double **a, double **b, double **c, int n, int bs)
{

  double **bt=matrix(n,n);
  matrix_transpose(b,bt,n);
  for (int i=0;i<n;i++)
    for (int j=0;j<n;j++)
      c[i][j]=0;

  int nb=n/bs;
  assert(nb*bs==n);  /*fail if we don't have an exact multiple of the block size*/
  
#pragma omp parallel default(none) shared(a,b,c,bt,n,bs,nb)   
  {
     double **myblock_1=matrix(bs,bs);
     double **myblock_2=matrix(bs,bs);
     double **myblock_3=matrix(bs,bs);
     #pragma omp for
     for (int ib=0;ib<nb;ib++)
        for (int jb=0;jb<nb;jb++)
           for (int kb=0;kb<nb;kb++)
           {
              int ii=ib*bs;
              int jj=jb*bs;
              int kk=kb*bs;
              for (int i=0;i<bs;i++)
                 for (int j=0;j<bs;j++)
                 {
                    myblock_1[i][j]=a[i+ii][j+kk];
                    myblock_2[i][j]=bt[i+jj][j+kk];
                 }
              for (int i=0;i<bs;i++)
                 for (int j=0;j<bs;j++)
                 {
                    myblock_3[i][j]=0;
                    for (int k=0;k<bs;k++)
                       myblock_3[i][j]+=myblock_1[i][k]*myblock_2[j][k];
                 }
              for (int i=0;i<bs;i++)
                 for (int j=0;j<bs;j++)
                    c[i+ii][j+jj]+=myblock_3[i][j];
           }
     free_matrix(myblock_1);
     free_matrix(myblock_2);
     free_matrix(myblock_3);
  }
  free_matrix(bt);    
}

/*=========================================================================*/
int main(int argc, char *argv[])
{
  pca_time tt;

  assert(argc>2); //make sure we passed in a size of the matrix
  int n=atoi(argv[1]);  /*get the size of the matrix*/
  int bs=atoi(argv[2]);
  int nb=n/bs;
  n=nb*bs;  /*make sure we have an exact multiple of the block size*/
  double **a=matrix(n,n);
  double **b=matrix(n,n);
  double **c=matrix(n,n);
  double **d=matrix(n,n);

  fill_random_matrix(a,n);
  fill_random_matrix(b,n);

  // as a reference:
  matmult_slowest(a,b,d,n);

  tick(&tt);
  matmult_slowest(a,b,c,n);
  fprintf(stdout,"Slowest version took %12.4e second. Error is %12.4e\n\n",
          tocksilent(&tt), find_matrix_error(c,d,n));

  tick(&tt);
  matmult_faster(a,b,d,n);
  fprintf(stdout,"Faster version took %12.4e second. Error is %12.4e\n\n",
          tocksilent(&tt), find_matrix_error(c,d,n));


  tick(&tt);
  matmult_block(a,b,d,n,bs);
  fprintf(stdout,"Block version took %12.4e second. Error is %12.4e\n\n",
          tocksilent(&tt), find_matrix_error(c,d,n));
  
  // tick(&tt);
  // matmult_blas(a,b,d,n);
  // fprintf(stdout,"Blas version took %12.4e seconds. Error is %12.4e\n\n",
  //         tocksilent(&tt), find_matrix_error(c,d,n));

}
